import datetime

import questdb_connect as qdbc
import sqlalchemy as sqla
from sqlalchemy.orm import Session

from tests.conftest import (
    ALL_TYPES_TABLE_NAME,
    METRICS_TABLE_NAME,
    collect_select_all,
    collect_select_all_raw_connection,
)


def test_insert(test_engine, test_model):
    with test_engine.connect() as conn:
        assert test_engine.dialect.has_table(conn, ALL_TYPES_TABLE_NAME)
        assert not test_engine.dialect.has_table(conn, 'scorchio')
        now = datetime.datetime(2023, 4, 12, 23, 55, 59, 342380)
        now_date = now.date()
        expected = ("(True, 8, 12, 13, 14, 15.234, 16.88993244, 'coconut', 'banana', 'C', "
                    "UUID('6d5eb038-63d1-4971-8484-30c16e13de5b'), datetime.datetime(2023, 4, 12, "
                    "0, 0), datetime.datetime(2023, 4, 12, 23, 55, 59, 342380), 'dfvgsj', "
                    "'0xa3b400fcf6ed707d710d5d4e672305203ed3cc6254d1cefe313e4a465861f42a')\n"
                    "(True, 8, 12, 13, 14, 15.234, 16.88993244, 'coconut', 'banana', 'C', "
                    "UUID('6d5eb038-63d1-4971-8484-30c16e13de5b'), datetime.datetime(2023, 4, 12, "
                    "0, 0), datetime.datetime(2023, 4, 12, 23, 55, 59, 342380), 'dfvgsj', "
                    "'0xa3b400fcf6ed707d710d5d4e672305203ed3cc6254d1cefe313e4a465861f42a')")
        insert_stmt = sqla.insert(test_model).values(
            col_boolean=True,
            col_byte=8,
            col_short=12,
            col_int=13,
            col_long=14,
            col_float=15.234,
            col_double=16.88993244,
            col_symbol='coconut',
            col_string='banana',
            col_char='C',
            col_uuid='6d5eb038-63d1-4971-8484-30c16e13de5b',
            col_date=now_date,
            col_ts=now,
            col_geohash='dfvgsj2vptwu',
            col_long256='0xa3b400fcf6ed707d710d5d4e672305203ed3cc6254d1cefe313e4a465861f42a'
        )
        conn.execute(insert_stmt)
        conn.execute(sqla.insert(test_model), {
            'col_boolean': True,
            'col_byte': 8,
            'col_short': 12,
            'col_int': 13,
            'col_long': 14,
            'col_float': 15.234,
            'col_double': 16.88993244,
            'col_symbol': 'coconut',
            'col_string': 'banana',
            'col_char': 'C',
            'col_uuid': '6d5eb038-63d1-4971-8484-30c16e13de5b',
            'col_date': now_date,
            'col_ts': now,
            'col_geohash': 'dfvgsj2vptwu',
            'col_long256': '0xa3b400fcf6ed707d710d5d4e672305203ed3cc6254d1cefe313e4a465861f42a'
        })
        assert collect_select_all(conn, expected_rows=2) == expected
    assert collect_select_all_raw_connection(test_engine, expected_rows=2) == expected


def test_inspect_1(test_engine, test_model):
    now = datetime.datetime(2023, 4, 12, 23, 55, 59, 342380)
    now_date = now.date()
    session = Session(test_engine)
    try:
        session.add(test_model(
            col_boolean=True,
            col_byte=8,
            col_short=12,
            col_int=0,
            col_long=14,
            col_float=15.234,
            col_double=16.88993244,
            col_symbol='coconut',
            col_string='banana',
            col_char='C',
            col_uuid='6d5eb038-63d1-4971-8484-30c16e13de5b',
            col_date=now_date,
            col_ts=now,
            col_geohash='dfvgsj2vptwu',
            col_long256='0xa3b400fcf6ed707d710d5d4e672305203ed3cc6254d1cefe313e4a465861f42a'
        ))
        session.commit()
    finally:
        if session:
            session.close()
    metadata = sqla.MetaData()
    table = sqla.Table(ALL_TYPES_TABLE_NAME, metadata, autoload_with=test_engine)
    table_columns = str([(col.name, col.type, col.primary_key) for col in table.columns])
    assert table_columns == str([
        ('col_boolean', qdbc.Boolean(), False),
        ('col_byte', qdbc.Byte(), False),
        ('col_short', qdbc.Short(), False),
        ('col_int', qdbc.Int(), False),
        ('col_long', qdbc.Long(), False),
        ('col_float', qdbc.Float(), False),
        ('col_double', qdbc.Double(), False),
        ('col_symbol', qdbc.Symbol(), False),
        ('col_string', qdbc.String(), False),
        ('col_char', qdbc.Char(), False),
        ('col_uuid', qdbc.UUID(), False),
        ('col_date', qdbc.Date(), False),
        ('col_ts', qdbc.Timestamp(), True),
        ('col_geohash', qdbc.GeohashInt(), False),
        ('col_long256', qdbc.Long256(), False)
    ])


def test_inspect_2(test_engine, test_metrics):
    metadata = sqla.MetaData()
    table = sqla.Table(METRICS_TABLE_NAME, metadata, autoload_with=test_engine)
    table_columns = str([(col.name, col.type, col.primary_key) for col in table.columns])
    assert table_columns == str([
        ('source', qdbc.Symbol(), False),
        ('attr_name', qdbc.Symbol(), False),
        ('attr_value', qdbc.Double(), False),
        ('ts', qdbc.Timestamp(), True),
    ])


def test_multiple_insert(test_engine, test_model):
    now = datetime.datetime(2023, 4, 12, 23, 55, 59, 342380)
    now_date = now.date()
    session = Session(test_engine)
    num_rows = 3
    expected = ("(True, 8, 12, 0, 14, 15.234, 16.88993244, 'coconut', 'banana', 'C', "
                "UUID('6d5eb038-63d1-4971-8484-30c16e13de5b'), datetime.datetime(2023, 4, 12, "
                "0, 0), datetime.datetime(2023, 4, 12, 23, 55, 59, 342380), 'dfvgsj', "
                "'0xa3b400fcf6ed707d710d5d4e672305203ed3cc6254d1cefe313e4a465861f42a')\n"
                "(True, 8, 12, 1, 14, 15.234, 16.88993244, 'coconut', 'banana', 'C', "
                "UUID('6d5eb038-63d1-4971-8484-30c16e13de5b'), datetime.datetime(2023, 4, 12, "
                "0, 0), datetime.datetime(2023, 4, 12, 23, 55, 59, 342380), 'dfvgsj', "
                "'0xa3b400fcf6ed707d710d5d4e672305203ed3cc6254d1cefe313e4a465861f42a')\n"
                "(True, 8, 12, 2, 14, 15.234, 16.88993244, 'coconut', 'banana', 'C', "
                "UUID('6d5eb038-63d1-4971-8484-30c16e13de5b'), datetime.datetime(2023, 4, 12, "
                "0, 0), datetime.datetime(2023, 4, 12, 23, 55, 59, 342380), 'dfvgsj', "
                "'0xa3b400fcf6ed707d710d5d4e672305203ed3cc6254d1cefe313e4a465861f42a')")
    try:
        for idx in range(num_rows):
            session.add(test_model(
                col_boolean=True,
                col_byte=8,
                col_short=12,
                col_int=idx,
                col_long=14,
                col_float=15.234,
                col_double=16.88993244,
                col_symbol='coconut',
                col_string='banana',
                col_char='C',
                col_uuid='6d5eb038-63d1-4971-8484-30c16e13de5b',
                col_date=now_date,
                col_ts=now,
                col_geohash='dfvgsj2vptwu',
                col_long256='0xa3b400fcf6ed707d710d5d4e672305203ed3cc6254d1cefe313e4a465861f42a'
            ))
            session.commit()
        assert collect_select_all(session, expected_rows=num_rows) == expected
    finally:
        if session:
            session.close()
        assert collect_select_all_raw_connection(test_engine, expected_rows=num_rows) == expected


def test_bulk_insert(test_engine, test_model):
    now = datetime.datetime(2023, 4, 12, 23, 55, 59, 342380)
    now_date = now.date()
    session = Session(test_engine)
    num_rows = 3
    expected = ("(True, 8, 12, 0, 14, 15.234, 16.88993244, 'coconut', 'banana', 'C', "
                "UUID('6d5eb038-63d1-4971-8484-30c16e13de5b'), datetime.datetime(2023, 4, 12, "
                "0, 0), datetime.datetime(2023, 4, 12, 23, 55, 59, 342380), 'dfvgsj', "
                "'0xa3b400fcf6ed707d710d5d4e672305203ed3cc6254d1cefe313e4a465861f42a')\n"
                "(True, 8, 12, 1, 14, 15.234, 16.88993244, 'coconut', 'banana', 'C', "
                "UUID('6d5eb038-63d1-4971-8484-30c16e13de5b'), datetime.datetime(2023, 4, 12, "
                "0, 0), datetime.datetime(2023, 4, 12, 23, 55, 59, 342380), 'dfvgsj', "
                "'0xa3b400fcf6ed707d710d5d4e672305203ed3cc6254d1cefe313e4a465861f42a')\n"
                "(True, 8, 12, 2, 14, 15.234, 16.88993244, 'coconut', 'banana', 'C', "
                "UUID('6d5eb038-63d1-4971-8484-30c16e13de5b'), datetime.datetime(2023, 4, 12, "
                "0, 0), datetime.datetime(2023, 4, 12, 23, 55, 59, 342380), 'dfvgsj', "
                "'0xa3b400fcf6ed707d710d5d4e672305203ed3cc6254d1cefe313e4a465861f42a')")
    models = [
        test_model(
            col_boolean=True,
            col_byte=8,
            col_short=12,
            col_int=idx,
            col_long=14,
            col_float=15.234,
            col_double=16.88993244,
            col_symbol='coconut',
            col_string='banana',
            col_char='C',
            col_uuid='6d5eb038-63d1-4971-8484-30c16e13de5b',
            col_date=now_date,
            col_ts=now,
            col_geohash='dfvgsj2vptwu',
            col_long256='0xa3b400fcf6ed707d710d5d4e672305203ed3cc6254d1cefe313e4a465861f42a'
        ) for idx in range(num_rows)
    ]
    try:
        session.bulk_save_objects(models)
        session.commit()
        assert collect_select_all(session, expected_rows=num_rows) == expected
    finally:
        if session:
            session.close()
        assert collect_select_all_raw_connection(test_engine, expected_rows=num_rows) == expected


def test_dialect_get_schema_names(test_engine):
    dialect = qdbc.QuestDBDialect()
    with test_engine.connect() as conn:
        assert dialect.get_schema_names(conn) == ["public"]


def test_dialect_get_table_names(test_engine):
    dialect = qdbc.QuestDBDialect()
    with test_engine.connect() as conn:
        table_names = dialect.get_table_names(conn, schema="public")
        assert table_names == dialect.get_table_names(conn)
        assert len(table_names) > 0


def test_dialect_has_table(test_engine):
    dialect = qdbc.QuestDBDialect()
    with test_engine.connect() as conn:
        for table_name in dialect.get_table_names(conn):
            if not dialect.has_table(conn, table_name):
                raise AssertionError()
            if not dialect.has_table(conn, table_name, schema="public"):
                raise AssertionError()


def test_functions(test_engine):
    with test_engine.connect() as conn:
        sql = sqla.text("SELECT name FROM functions()")
        expected = [row[0] for row in conn.execute(sql).fetchall()]
        assert qdbc.get_functions_list() == expected


def test_keywords(test_engine):
    with test_engine.connect() as conn:
        sql = sqla.text("SELECT keyword FROM keywords()")
        expected = [row[0] for row in conn.execute(sql).fetchall()]
        assert qdbc.get_keywords_list() == expected



import typing

import sqlalchemy

from .common import PartitionBy, quote_identifier


class QDBTableEngine(
    sqlalchemy.sql.base.SchemaEventTarget, sqlalchemy.sql.visitors.Traversible
):
    def __init__(
        self,
        table_name: str,
        ts_col_name: str,
        partition_by: PartitionBy = PartitionBy.DAY,
        is_wal: bool = True,
        dedup_upsert_keys: typing.Optional[typing.Tuple[str]] = None,
    ):
        sqlalchemy.sql.visitors.Traversible.__init__(self)
        self.name = table_name
        self.ts_col_name = ts_col_name
        self.partition_by = partition_by
        self.is_wal = is_wal
        self.dedup_upsert_keys = dedup_upsert_keys
        self.compiled = None

    def get_table_suffix(self):
        if self.compiled is None:
            self.compiled = ""
            has_ts = self.ts_col_name is not None
            is_partitioned = self.partition_by and self.partition_by != PartitionBy.NONE
            if has_ts:
                self.compiled += f'TIMESTAMP("{self.ts_col_name}")'
            if is_partitioned:
                if not has_ts:
                    raise sqlalchemy.exc.ArgumentError(
                        None,
                        "Designated timestamp must be specified for partitioned table",
                    )
                self.compiled += f" PARTITION BY {self.partition_by.name}"
            if self.is_wal:
                if not is_partitioned:
                    raise sqlalchemy.exc.ArgumentError(
                        None, "WAL table requires designated timestamp and partition by"
                    )
                if self.is_wal:
                    self.compiled += " WAL"
                    if self.dedup_upsert_keys:
                        self.compiled += " DEDUP UPSERT KEYS("
                        self.compiled += ",".join(
                            map(quote_identifier, self.dedup_upsert_keys)
                        )
                        self.compiled += ")"
                else:
                    if self.dedup_upsert_keys:
                        raise sqlalchemy.exc.ArgumentError(
                            None, "DEDUP only applies to WAL tables"
                        )
                    self.compiled += " BYPASS WAL"
        return self.compiled

    def _set_parent(self, parent, **_kwargs):
        parent.engine = self



import abc

import sqlalchemy

from .common import quote_identifier


def _none(_ignore):
    return None


_special_chars = {
    "(",
    ")",
    "[",
    "[]",
    "{",
    "}",
    "'",
    '"',
    ":",
    ";",
    ".",
    "!",
    "%",
    "&",
    "*",
    "$",
    "@",
    "~",
    "^",
    "-",
    "?",
    "/",
    "\\",
    " ",
    "\t",
    "\r",
    "\n",
}


def _has_special_char(_value):
    for candidate in _value:
        if candidate in _special_chars:
            return True
    return False


class QDBIdentifierPreparer(sqlalchemy.sql.compiler.IdentifierPreparer, abc.ABC):
    schema_for_object = staticmethod(_none)

    def __init__(
        self,
        dialect,
        initial_quote='"',
        final_quote=None,
        escape_quote='"',
        quote_case_sensitive_collations=False,
        omit_schema=True,
    ):
        super().__init__(
            dialect=dialect,
            initial_quote=initial_quote,
            final_quote=final_quote,
            escape_quote=escape_quote,
            quote_case_sensitive_collations=quote_case_sensitive_collations,
            omit_schema=omit_schema,
        )

    def quote_identifier(self, value):
        return quote_identifier(value)

    def _requires_quotes(self, _value):
        return _value and _has_special_char(_value)

    def format_schema(self, name):
        """Prepare a quoted schema name."""
        return ""

    def format_table(self, table, use_schema=True, name=None):
        """Prepare a quoted table and schema name."""
        return quote_identifier(name if name else table.name)



import abc

import psycopg2
import sqlalchemy

from .common import PartitionBy
from .table_engine import QDBTableEngine
from .types import resolve_type_from_name


class QDBInspector(sqlalchemy.engine.reflection.Inspector, abc.ABC):
    def reflecttable(
            self,
            table,
            include_columns,
            exclude_columns=(),
            resolve_fks=True,
            _extend_on=None,
    ):
        # backward compatibility SQLAlchemy 1.3
        return self.reflect_table(
            table, include_columns, exclude_columns, resolve_fks, _extend_on
        )

    def reflect_table(
            self,
            table,
            include_columns=None,
            exclude_columns=None,
            resolve_fks=False,
            _extend_on=None,
            _reflect_info=None
    ):
        table_name = table.name
        try:
            result_set = self.bind.execute(
                sqlalchemy.text(
                    "SELECT designatedTimestamp, partitionBy, walEnabled FROM tables() WHERE table_name = :tn"), {"tn": table_name}
            )
        except psycopg2.DatabaseError:
            # older version
            result_set = self.bind.execute(
                sqlalchemy.text(
                    "SELECT designatedTimestamp, partitionBy, walEnabled FROM tables() WHERE name = :tn"), {"tn": table_name}
            )
        if not result_set:
            self._panic_table(table_name)
        table_attrs = result_set.first()
        if table_attrs:
            col_ts_name = table_attrs[0]
            partition_by = PartitionBy[table_attrs[1]]
            is_wal = True if table_attrs[2] else False
        else:
            col_ts_name = None
            partition_by = PartitionBy.NONE
            is_wal = True
        dedup_upsert_keys = []
        for row in self.bind.execute(
                sqlalchemy.text("SELECT \"column\", \"type\", \"upsertKey\" FROM table_columns(:tn)"), {"tn": table_name}
        ):
            col_name = row[0]
            if include_columns and col_name not in include_columns:
                continue
            if exclude_columns and col_name in exclude_columns:
                continue
            if row[2]:  # upsertKey
                dedup_upsert_keys.append(col_name)
            col_type = resolve_type_from_name(row[1])
            table.append_column(
                sqlalchemy.Column(col_name, col_type, primary_key=(col_ts_name and col_ts_name.upper() == col_name.upper()))
            )
        table.engine = QDBTableEngine(
            table_name,
            col_ts_name,
            partition_by,
            is_wal,
            tuple(dedup_upsert_keys) if dedup_upsert_keys else None,
        )
        table.metadata = sqlalchemy.MetaData()

    def get_columns(self, table_name, schema=None, **kw):
        result_set = self.bind.execute(sqlalchemy.text("SELECT \"column\", \"type\" FROM table_columns(:tn)"), {"tn": table_name})
        return self.format_table_columns(table_name, result_set)

    def get_schema_names(self):
        return ["public"]

    def format_table_columns(self, table_name, result_set):
        if not result_set:
            self._panic_table(table_name)
        return [
            {
                "name": row[0],
                "type": resolve_type_from_name(row[1])(),
                "nullable": True,
                "autoincrement": False,
            }
            for row in result_set
        ]

    def _panic_table(self, table_name):
        raise sqlalchemy.orm.exc.NoResultFound(f"Table '{table_name}' does not exist")



import sqlalchemy as sqla


def main():
    print(f"SqlAlchemy {sqla.__version__}")

    table_name = "sqlalchemy_raw"
    engine = sqla.create_engine("questdb://localhost:8812/main", echo=True, future=True)
    try:
        with engine.connect() as conn:
            conn.execute(sqla.text(f"DROP TABLE IF EXISTS {table_name}"))
            conn.execute(
                sqla.text(f"CREATE TABLE IF NOT EXISTS {table_name} (x int, y int)")
            )
            conn.execute(
                sqla.text(f"INSERT INTO {table_name} (x, y) VALUES (:x, :y)"),
                [{"x": 1, "y": 1}, {"x": 2, "y": 4}],
            )
            conn.commit()

            result = conn.execute(
                sqla.text(f"SELECT x, y FROM {table_name} WHERE y > :y"), {"y": 2}
            )
            for row in result:
                print(f"x: {row.x}  y: {row.y}")

            result = conn.execute(sqla.text(f"SELECT x, y FROM {table_name}"))
            for dict_row in result.mappings():
                x = dict_row["x"]
                y = dict_row["y"]
                print(f"x: {x}  y: {y}")
    finally:
        if engine:
            engine.dispose()


if __name__ == "__main__":
    main()



import abc

import sqlalchemy

from .common import quote_identifier, remove_public_schema
from .types import QDBTypeMixin


class QDBDDLCompiler(sqlalchemy.sql.compiler.DDLCompiler, abc.ABC):
    def visit_create_schema(self, create, **kw):
        raise Exception("QuestDB does not support SCHEMAS, there is only 'public'")

    def visit_drop_schema(self, drop, **kw):
        raise Exception("QuestDB does not support SCHEMAS, there is only 'public'")

    def visit_create_table(self, create, **kw):
        table = create.element
        create_table = f"CREATE TABLE {quote_identifier(table.fullname)} ("
        create_table += ", ".join(
            [self.get_column_specification(c.element) for c in create.columns]
        )
        return create_table + ") " + table.engine.get_table_suffix()

    def get_column_specification(self, column: sqlalchemy.Column, **_):
        if not isinstance(column.type, QDBTypeMixin):
            raise sqlalchemy.exc.ArgumentError(
                "Column type is not a valid QuestDB type"
            )
        return column.type.column_spec(column.name)


class QDBSQLCompiler(sqlalchemy.sql.compiler.SQLCompiler, abc.ABC):
    def _is_safe_for_fast_insert_values_helper(self):
        return True

    def visit_textclause(self, textclause, add_to_result_map=None, **kw):
        textclause.text = remove_public_schema(textclause.text)
        return super().visit_textclause(textclause, add_to_result_map, **kw)



import os

CONNECTION_ATTRS = {
    "host": os.environ.get("QUESTDB_CONNECT_HOST", "localhost"),
    "port": int(os.environ.get("QUESTDB_CONNECT_PORT", "8812")),
    "username": os.environ.get("QUESTDB_CONNECT_USER", "admin"),
    "password": os.environ.get("QUESTDB_CONNECT_PASSWORD", "quest"),
}



"""
This code connects to a running QuestDB instance and displays information
about the tables and partitions in the database. It uses two SQL statements
to retrieve this information: "tables()" and "table_partitions()". The output
is formatted and printed to the console.
"""
import json

from questdb_connect import connect

from examples import CONNECTION_ATTRS


def print_partition(row):
    p_index, p_by, _, min_ts, max_ts, num_rows, _, p_size, *_ = row
    print(
        f" - Partition {p_index} by {p_by} [{min_ts}, {max_ts}] {num_rows} rows {p_size}"
    )


def print_table(row):
    table_id, table_name, ts_column, p_by, _, _, is_wal, dir_name, is_dedup = row
    msg = ", ".join(
        (
            f"Table id:{table_id}",
            f"name:{table_name}",
            f"ts-col:{ts_column}",
            f"partition-by:{p_by}",
            f"is-wal:{is_wal}",
            f"dir-name:{dir_name}",
            f"is-dedup:{is_dedup}",
        )
    )
    print(msg)


def print_server_info(dsn_parameters):
    print(f"QuestDB server information: {json.dumps(dsn_parameters, indent=4)}")


def main():
    with connect(**CONNECTION_ATTRS) as conn:
        print_server_info(conn.get_dsn_parameters())
        with conn.cursor() as tables_cur, conn.cursor() as partitions_cur:
            tables_cur.execute("tables()")
            for table_row in tables_cur.fetchall():
                print_table(table_row)
                partitions_cur.execute(f"table_partitions('{table_row[1]}')")
                for partition_row in partitions_cur.fetchall():
                    print_partition(partition_row)


if __name__ == "__main__":
    main()



import re

import questdb_connect as qdbc


def test_resolve_type_from_name():
    for type_class in qdbc.QUESTDB_TYPES:
        resolved_class = qdbc.resolve_type_from_name(type_class.__visit_name__)
        assert type_class.__visit_name__ == resolved_class.__visit_name__
        assert isinstance(type_class(), resolved_class)
        assert isinstance(resolved_class(), type_class)

    for n in range(1, 61):
        g_name = qdbc.geohash_type_name(n)
        g_class = qdbc.resolve_type_from_name(g_name)
        assert isinstance(g_class(), qdbc.geohash_class(n))


def test_superset_default_mappings():
    default_column_type_mappings = (
        (re.compile("^BOOLEAN$", re.IGNORECASE), qdbc.Boolean),
        (re.compile("^BYTE$", re.IGNORECASE), qdbc.Byte),
        (re.compile("^SHORT$", re.IGNORECASE), qdbc.Short),
        (re.compile("^CHAR$", re.IGNORECASE), qdbc.Char),
        (re.compile("^INT$", re.IGNORECASE), qdbc.Int),
        (re.compile("^LONG$", re.IGNORECASE), qdbc.Long),
        (re.compile("^DATE$", re.IGNORECASE), qdbc.Date),
        (re.compile("^TIMESTAMP$", re.IGNORECASE), qdbc.Timestamp),
        (re.compile("^FLOAT$", re.IGNORECASE), qdbc.Float),
        (re.compile("^DOUBLE$", re.IGNORECASE), qdbc.Double),
        (re.compile("^STRING$", re.IGNORECASE), qdbc.String),
        (re.compile("^SYMBOL$", re.IGNORECASE), qdbc.Symbol),
        (re.compile("^LONG256$", re.IGNORECASE), qdbc.Long256),
        (re.compile("^UUID$", re.IGNORECASE), qdbc.UUID),
        (re.compile("^LONG118$", re.IGNORECASE), qdbc.UUID),
        (re.compile("^IPV4$", re.IGNORECASE), qdbc.IPv4),
    )
    for type_class in qdbc.QUESTDB_TYPES:
        for pattern, _expected_type in default_column_type_mappings:
            matching_name = pattern.match(type_class.__visit_name__)
            if matching_name:
                print(f"match: {matching_name}, type_class: {type_class}")
                resolved_class = qdbc.resolve_type_from_name(matching_name.group(0))
                assert type_class.__visit_name__ == resolved_class.__visit_name__
                assert isinstance(type_class(), resolved_class)
                assert isinstance(resolved_class(), type_class)
                break
    geohash_pattern = re.compile(r"^GEOHASH\(\d+[b|c]\)$", re.IGNORECASE)
    for n in range(1, 61):
        g_name = qdbc.geohash_type_name(n)
        matching_name = geohash_pattern.match(g_name).group(0)
        assert matching_name == g_name
        g_class = qdbc.resolve_type_from_name(g_name)
        assert isinstance(g_class(), qdbc.geohash_class(n))



from examples import hello_world, psycopg2_connect, server_utilisation, sqlalchemy_orm, sqlalchemy_raw


def test_hello_world():
    hello_world.main()


def test_psycopg2_connect():
    psycopg2_connect.main()


def test_server_utilisation():
    server_utilisation.main(duration_sec=2.0)


def test_sqlalchemy_orm():
    sqlalchemy_orm.main()


def test_sqlalchemy_raw():
    sqlalchemy_raw.main()



def get_keywords_list(conn=None):
    return __initialize_list(
        conn, "SELECT keyword FROM keywords()", __keywords, __default_keywords
    )


def get_functions_list(conn=None):
    return __initialize_list(
        conn, "SELECT name FROM functions()", __func_names, __default_func_names
    )


def __initialize_list(conn, sql_stmt, target_list, default_target_list):
    if not target_list:
        try:
            with conn.cursor() as functions_cur:
                functions_cur.execute(sql_stmt)
                for func_row in functions_cur.fetchall():
                    target_list.append(func_row[0])
        except Exception as _ignore:
            target_list.extend(default_target_list)
    return target_list


__func_names = []
__default_func_names = [
    "abs",
    "acos",
    "all_tables",
    "and",
    "asin",
    "atan",
    "atan2",
    "avg",
    "base64",
    "between",
    "build",
    "case",
    "cast",
    "ceil",
    "ceiling",
    "coalesce",
    "concat",
    "cos",
    "cot",
    "count",
    "count_distinct",
    "current_database",
    "current_schema",
    "current_schemas",
    "current_user",
    "date_trunc",
    "dateadd",
    "datediff",
    "day",
    "day_of_week",
    "day_of_week_sunday_first",
    "days_in_month",
    "degrees",
    "dump_memory_usage",
    "dump_thread_stacks",
    "extract",
    "first",
    "floor",
    "flush_query_cache",
    "format_type",
    "haversine_dist_deg",
    "hour",
    "ilike",
    "information_schema._pg_expandarray",
    "isOrdered",
    "is_leap_year",
    "ksum",
    "last",
    "left",
    "length",
    "like",
    "list",
    "log",
    "long_sequence",
    "lower",
    "lpad",
    "ltrim",
    "make_geohash",
    "max",
    "memory_metrics",
    "micros",
    "millis",
    "min",
    "minute",
    "month",
    "not",
    "now",
    "nsum",
    "nullif",
    "pg_advisory_unlock_all",
    "pg_attrdef",
    "pg_attribute",
    "pg_catalog.age",
    "pg_catalog.current_database",
    "pg_catalog.current_schema",
    "pg_catalog.current_schemas",
    "pg_catalog.pg_attrdef",
    "pg_catalog.pg_attribute",
    "pg_catalog.pg_class",
    "pg_catalog.pg_database",
    "pg_catalog.pg_description",
    "pg_catalog.pg_get_expr",
    "pg_catalog.pg_get_keywords",
    "pg_catalog.pg_get_partkeydef",
    "pg_catalog.pg_get_userbyid",
    "pg_catalog.pg_index",
    "pg_catalog.pg_inherits",
    "pg_catalog.pg_is_in_recovery",
    "pg_catalog.pg_locks",
    "pg_catalog.pg_namespace",
    "pg_catalog.pg_roles",
    "pg_catalog.pg_shdescription",
    "pg_catalog.pg_table_is_visible",
    "pg_catalog.pg_type",
    "pg_catalog.txid_current",
    "pg_catalog.version",
    "pg_class",
    "pg_database",
    "pg_description",
    "pg_get_expr",
    "pg_get_keywords",
    "pg_get_partkeydef",
    "pg_index",
    "pg_inherits",
    "pg_is_in_recovery",
    "pg_locks",
    "pg_namespace",
    "pg_postmaster_start_time",
    "pg_proc",
    "pg_range",
    "pg_roles",
    "pg_type",
    "position",
    "power",
    "radians",
    "reader_pool",
    "regexp_replace",
    "replace",
    "right",
    "rnd_bin",
    "rnd_boolean",
    "rnd_byte",
    "rnd_char",
    "rnd_date",
    "rnd_double",
    "rnd_float",
    "rnd_geohash",
    "rnd_int",
    "rnd_log",
    "rnd_long",
    "rnd_long256",
    "rnd_short",
    "rnd_str",
    "rnd_symbol",
    "rnd_timestamp",
    "rnd_uuid4",
    "round",
    "round_down",
    "round_half_even",
    "round_up",
    "row_number",
    "rpad",
    "rtrim",
    "second",
    "session_user",
    "simulate_crash",
    "sin",
    "size_pretty",
    "split_part",
    "sqrt",
    "starts_with",
    "stddev_samp",
    "string_agg",
    "strpos",
    "substring",
    "sum",
    "switch",
    "sysdate",
    "systimestamp",
    "table_columns",
    "table_partitions",
    "table_writer_metrics",
    "tables",
    "tan",
    "timestamp_ceil",
    "timestamp_floor",
    "timestamp_sequence",
    "timestamp_shuffle",
    "to_char",
    "to_date",
    "to_long128",
    "to_lowercase",
    "to_pg_date",
    "to_str",
    "to_timestamp",
    "to_timezone",
    "to_uppercase",
    "to_utc",
    "touch",
    "trim",
    "txid_current",
    "typeOf",
    "upper",
    "version",
    "wal_tables",
    "week_of_year",
    "year",
]
__keywords = []
__default_keywords = [
    "add",
    "all",
    "alter",
    "and",
    "as",
    "asc",
    "asof",
    "backup",
    "between",
    "by",
    "cache",
    "capacity",
    "case",
    "cast",
    "column",
    "columns",
    "copy",
    "create",
    "cross",
    "database",
    "default",
    "delete",
    "desc",
    "distinct",
    "drop",
    "else",
    "end",
    "except",
    "exists",
    "fill",
    "foreign",
    "from",
    "grant",
    "group",
    "header",
    "if",
    "in",
    "index",
    "inner",
    "insert",
    "intersect",
    "into",
    "isolation",
    "join",
    "key",
    "latest",
    "limit",
    "lock",
    "lt",
    "nan",
    "natural",
    "nocache",
    "none",
    "not",
    "null",
    "on",
    "only",
    "or",
    "order",
    "outer",
    "over",
    "partition",
    "primary",
    "references",
    "rename",
    "repair",
    "right",
    "sample",
    "select",
    "show",
    "splice",
    "system",
    "table",
    "tables",
    "then",
    "to",
    "transaction",
    "truncate",
    "type",
    "union",
    "unlock",
    "update",
    "values",
    "when",
    "where",
    "with",
    "writer",
]



import enum
import re


class PartitionBy(enum.Enum):
    DAY = 0
    MONTH = 1
    YEAR = 2
    NONE = 3
    HOUR = 4
    WEEK = 5


def remove_public_schema(query):
    try:
        if isinstance(query, str) and query is not None and "public" in query:
            return re.sub(_PUBLIC_SCHEMA_FILTER, "", query)
    except Exception:
        return query
    return query


def quote_identifier(identifier: str):
    if not identifier:
        return None
    first = 0
    last = len(identifier)
    if identifier[first] in _QUOTES:
        first += 1
    if identifier[last - 1] in _QUOTES:
        last -= 1
    return f'"{identifier[first:last]}"'


_PUBLIC_SCHEMA_FILTER = re.compile(
    r"(')?(public(?(1)\1|)\.)", re.IGNORECASE | re.MULTILINE
)
_QUOTES = ("'", '"')



import datetime
import os

os.environ.setdefault("SQLALCHEMY_SILENCE_UBER_WARNING", "1")

import questdb_connect as qdbc
from sqlalchemy import Column, MetaData, create_engine, insert
from sqlalchemy.orm import declarative_base

Base = declarative_base(metadata=MetaData())


class Signal(Base):
    __tablename__ = "signal"
    __table_args__ = (
        qdbc.QDBTableEngine("signal", "ts", qdbc.PartitionBy.HOUR, is_wal=True),
    )
    source = Column(qdbc.Symbol)
    value = Column(qdbc.Double)
    ts = Column(qdbc.Timestamp, primary_key=True)


def main():
    engine = create_engine("questdb://localhost:8812/main")
    try:
        Base.metadata.create_all(engine)
        with engine.connect() as conn:
            conn.execute(
                insert(Signal).values(
                    source="coconut", value=16.88993244, ts=datetime.datetime.utcnow()
                )
            )
    finally:
        if engine:
            engine.dispose()


if __name__ == "__main__":
    main()



import datetime
from unittest import mock

import pytest
from qdb_superset.db_engine_specs.questdb import QuestDbEngineSpec
from questdb_connect.types import QUESTDB_TYPES, Timestamp
from sqlalchemy import column, literal_column
from sqlalchemy.types import TypeEngine


def test_build_sqlalchemy_uri():
    assert QuestDbEngineSpec.build_sqlalchemy_uri(
        {
            "host": "localhost",
            "port": "8812",
            "username": "admin",
            "password": "quest",
            "database": "main",
        }
    ) == "questdb://admin:quest@localhost:8812/main"


def test_default_schema_for_query():
    assert QuestDbEngineSpec.get_default_schema_for_query("main", None) == None


def test_get_text_clause():
    sql_clause = "SELECT * FROM public.mytable t1"
    sql_clause += " JOIN public.myclient t2 ON t1.id = t2.id"
    expected_clause = "SELECT * FROM mytable t1 JOIN myclient t2 ON t1.id = t2.id"
    actual_clause = str(QuestDbEngineSpec.get_text_clause(sql_clause))
    print(f"sql: {sql_clause}, ex: {expected_clause}, ac: {actual_clause}")
    assert expected_clause == actual_clause


def test_epoch_to_dttm():
    assert QuestDbEngineSpec.epoch_to_dttm() == "{col} * 1000000"


@pytest.mark.parametrize(
    ("target_type", "expected_result", "dttm"),
    [
        (
                "Date",
                "TO_DATE('2023-04-28', 'YYYY-MM-DD')",
                datetime.datetime(2023, 4, 28, 23, 55, 59, 281567),
        ),
        (
                "DateTime",
                "TO_TIMESTAMP('2023-04-28T23:55:59.281567', 'yyyy-MM-ddTHH:mm:ss.SSSUUU')",
                datetime.datetime(2023, 4, 28, 23, 55, 59, 281567),
        ),
        (
                "TimeStamp",
                "TO_TIMESTAMP('2023-04-28T23:55:59.281567', 'yyyy-MM-ddTHH:mm:ss.SSSUUU')",
                datetime.datetime(2023, 4, 28, 23, 55, 59, 281567),
        ),
        ("UnknownType", None, datetime.datetime(2023, 4, 28, 23, 55, 59, 281567)),
    ],
)
def test_convert_dttm(target_type, expected_result, dttm) -> None:
    # datetime(year, month, day, hour, minute, second, microsecond)
    for target in (
            target_type,
            target_type.upper(),
            target_type.lower(),
            target_type.capitalize(),
    ):
        assert QuestDbEngineSpec.convert_dttm(
            target_type=target, dttm=dttm
        ) == expected_result


def test_get_datatype():
    assert QuestDbEngineSpec.get_datatype("int") == "INT"
    assert QuestDbEngineSpec.get_datatype(["int"]) == "['int']"


def test_get_column_spec():
    for native_type in QUESTDB_TYPES:
        column_spec = QuestDbEngineSpec.get_column_spec(native_type.__visit_name__)
        assert native_type == column_spec.sqla_type
        assert native_type != Timestamp or column_spec.is_dttm


def test_get_sqla_column_type():
    for native_type in QUESTDB_TYPES:
        column_type = QuestDbEngineSpec.get_sqla_column_type(native_type.__visit_name__)
        assert isinstance(column_type, TypeEngine.__class__)


def test_get_allow_cost_estimate():
    assert not QuestDbEngineSpec.get_allow_cost_estimate(extra=None)


def test_get_view_names():
    assert set() == QuestDbEngineSpec.get_view_names("main", None, None)


def test_get_table_names():
    inspector = mock.Mock()
    inspector.get_table_names = mock.Mock(
        return_value=["public.table", "table_2", '"public.table_3"']
    )
    pg_result = QuestDbEngineSpec.get_table_names(
        database=mock.ANY, schema="public", inspector=inspector
    )
    assert {"table", '"public.table_3"', "table_2"} == pg_result


def test_time_exp_literal_no_grain(test_engine):
    col = literal_column("COALESCE(a, b)")
    expr = QuestDbEngineSpec.get_timestamp_expr(col, None, None)
    result = str(expr.compile(None, dialect=test_engine.dialect))
    assert "COALESCE(a, b)" == result


def test_time_ex_lowr_col_no_grain(test_engine):
    col = column("lower_case")
    expr = QuestDbEngineSpec.get_timestamp_expr(col, None, None)
    result = str(expr.compile(None, dialect=test_engine.dialect))
    assert "lower_case" == result


def test_execute_sql_statement(test_engine) -> None:
    query = """
        SELECT timestamp, country_code, SUM(load_forecast) AS daily_forecast,
           SUM(load_actual) AS daily_actual
        FROM public.energy_2018
        SAMPLE BY 1d ALIGN TO CALENDAR;
        """
    with test_engine.connect() as cursor:
        rs = QuestDbEngineSpec.execute(cursor, query)
        print (rs)



import pytest
from _pytest.outcomes import fail
from psycopg2 import OperationalError
from sqlalchemy import create_engine


def test_user(test_engine, test_model):
    engine = create_engine("questdb://admin:quest@localhost:8812/qdb")
    engine.connect()

    engine = create_engine("questdb://user1:quest@localhost:8812/qdb")
    with pytest.raises(OperationalError) as exc_info:
        engine.connect()
    if not str(exc_info.value).__contains__("ERROR:  invalid username/password"):
        fail()



from __future__ import annotations

import re
from datetime import datetime
from typing import Any

import questdb_connect.types as qdbc_types
from flask_babel import gettext as __
from marshmallow import fields, Schema
from questdb_connect.common import remove_public_schema
from sqlalchemy.engine.base import Engine
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.sql.expression import text, TextClause
from sqlalchemy.types import TypeEngine
from sqlalchemy import text
import logging

# Configure the logging
logging.basicConfig(level=logging.ERROR)
logger = logging.getLogger(__name__)

from superset.db_engine_specs.base import (
    BaseEngineSpec,
    BasicParametersMixin,
    BasicParametersType,
)
from superset import sql_parse
from superset.utils import core as utils
from superset.utils.core import GenericDataType


class QuestDbParametersSchema(Schema):
    username = fields.String(
        metadata={"description": __("username")},
        dump_default="admin",
        load_default="admin",
    )
    password = fields.String(
        metadata={"description": __("password")},
        dump_default="quest",
        load_default="quest",
    )
    host = fields.String(
        metadata={"description": __("host")},
        dump_default="host.docker.internal",
        load_default="host.docker.internal",
    )
    port = fields.Integer(
        metadata={"description": __("port")},
        dump_default="8812",
        load_default="8812",
    )
    database = fields.String(
        metadata={"description": __("database")},
        dump_default="main",
        load_default="main",
    )


class QuestDbEngineSpec(BaseEngineSpec, BasicParametersMixin):
    engine = "questdb"
    engine_name = "QuestDB"
    default_driver = "psycopg2"
    encryption_parameters = {"sslmode": "prefer"}
    sqlalchemy_uri_placeholder = "questdb://username:password@host:port/database"
    parameters_schema = QuestDbParametersSchema()
    time_groupby_inline = False
    allows_hidden_cc_in_orderby = True
    time_secondary_columns = True
    try_remove_schema_from_table_name = True
    max_column_name_length = 120
    supports_dynamic_schema = False
    top_keywords: set[str] = set({})
    # https://en.wikipedia.org/wiki/ISO_8601#Durations
    # https://questdb.io/docs/reference/function/date-time/#date_trunc
    _time_grain_expressions = {
        None: "{col}",
        "PT1S": "DATE_TRUNC('second', {col})",
        "PT1M": "DATE_TRUNC('minute', {col})",
        "PT1H": "DATE_TRUNC('hour', {col})",
        "P1D": "DATE_TRUNC('day', {col})",
        "P1W": "DATE_TRUNC('week', {col})",
        "P1M": "DATE_TRUNC('month', {col})",
        "P1Y": "DATE_TRUNC('year', {col})",
        "P3M": "DATE_TRUNC('quarter', {col})",
    }
    column_type_mappings = (
        (
            re.compile("^BOOLEAN$", re.IGNORECASE),
            qdbc_types.Boolean,
            GenericDataType.BOOLEAN,
        ),
        (re.compile("^BYTE$", re.IGNORECASE), qdbc_types.Byte, GenericDataType.NUMERIC),
        (
            re.compile("^SHORT$", re.IGNORECASE),
            qdbc_types.Short,
            GenericDataType.NUMERIC,
        ),
        (re.compile("^CHAR$", re.IGNORECASE), qdbc_types.Char, GenericDataType.STRING),
        (re.compile("^INT$", re.IGNORECASE), qdbc_types.Int, GenericDataType.NUMERIC),
        (re.compile("^LONG$", re.IGNORECASE), qdbc_types.Long, GenericDataType.NUMERIC),
        (
            re.compile("^DATE$", re.IGNORECASE),
            qdbc_types.Date,
            GenericDataType.TEMPORAL,
        ),
        (
            re.compile("^TIMESTAMP$", re.IGNORECASE),
            qdbc_types.Timestamp,
            GenericDataType.TEMPORAL,
        ),
        (
            re.compile("^FLOAT$", re.IGNORECASE),
            qdbc_types.Float,
            GenericDataType.NUMERIC,
        ),
        (
            re.compile("^DOUBLE$", re.IGNORECASE),
            qdbc_types.Double,
            GenericDataType.NUMERIC,
        ),
        (
            re.compile("^STRING$", re.IGNORECASE),
            qdbc_types.String,
            GenericDataType.STRING,
        ),
        (
            re.compile("^SYMBOL$", re.IGNORECASE),
            qdbc_types.Symbol,
            GenericDataType.STRING,
        ),
        (
            re.compile("^LONG256$", re.IGNORECASE),
            qdbc_types.Long256,
            GenericDataType.STRING,
        ),
        (
            re.compile(r"^GEOHASH\(\d+[b|c]\)$", re.IGNORECASE),
            qdbc_types.GeohashLong,
            GenericDataType.STRING,
        ),
        (re.compile("^UUID$", re.IGNORECASE), qdbc_types.UUID, GenericDataType.STRING),
        (
            re.compile("^LONG128$", re.IGNORECASE),
            qdbc_types.Long128,
            GenericDataType.STRING,
        ),
        (re.compile("^IPV4$", re.IGNORECASE), qdbc_types.IPv4, GenericDataType.STRING),
    )

    @classmethod
    def build_sqlalchemy_uri(
        cls,
        parameters: BasicParametersType,
        encrypted_extra: dict[str, str] | None = None,
    ) -> str:
        host = parameters.get("host")
        port = parameters.get("port")
        username = parameters.get("username")
        password = parameters.get("password")
        database = parameters.get("database")
        return f"questdb://{username}:{password}@{host}:{port}/{database}"

    @classmethod
    def get_default_schema_for_query(cls, database, query) -> str | None:
        """Return the default schema for a given query."""
        return None

    @classmethod
    def epoch_to_dttm(cls) -> str:
        """SQL expression that converts epoch (seconds) to datetime that can be used
        in a query. The reference column should be denoted as `{col}` in the return
        expression, e.g. "FROM_UNIXTIME({col})"
        :return: SQL Expression
        """
        return "{col} * 1000000"

    @classmethod
    def convert_dttm(
        cls, target_type: str, dttm: datetime, db_extra: dict[str, Any] | None = None
    ) -> str | None:
        """Convert a Python `datetime` object to a SQL expression.
        :param target_type: The target type of expression
        :param dttm: The datetime object
        :return: The SQL expression
        """
        type_u = target_type.upper()
        if type_u == "DATE":
            return f"TO_DATE('{dttm.date().isoformat()}', 'YYYY-MM-DD')"
        if type_u in ("DATETIME", "TIMESTAMP"):
            dttm_formatted = dttm.isoformat(sep="T", timespec="microseconds")
            return f"TO_TIMESTAMP('{dttm_formatted}', 'yyyy-MM-ddTHH:mm:ss.SSSUUU')"
        return None

    @classmethod
    def get_datatype(cls, type_code: Any) -> str | None:
        """Change column type code from cursor description to string representation.
        :param type_code: Type code from cursor description
        :return: String representation of type code
        """
        if isinstance(type_code, str) and type_code:
            return type_code.upper()
        return str(type_code)

    @classmethod
    def get_column_spec(
        cls,
        native_type: str | None,
        db_extra: dict[str, Any] | None = None,
        source: utils.ColumnTypeSource = utils.ColumnTypeSource.GET_TABLE,
    ) -> utils.ColumnSpec | None:
        """Get generic type related specs regarding a native column type.
        :param native_type: Native database type
        :param db_extra: The database extra object
        :param source: Type coming from the database table or cursor description
        :return: ColumnSpec object
        """
        sqla_type = qdbc_types.resolve_type_from_name(native_type)
        if not sqla_type:
            return BaseEngineSpec.get_column_spec(native_type, db_extra, source)
        name_u = sqla_type.__visit_name__
        generic_type = GenericDataType.STRING
        if name_u == "BOOLEAN":
            generic_type = GenericDataType.BOOLEAN
        elif name_u in ("BYTE", "SHORT", "INT", "LONG", "FLOAT", "DOUBLE"):
            generic_type = GenericDataType.NUMERIC
        elif name_u in (
            "SYMBOL",
            "STRING",
            "CHAR",
            "LONG256",
            "UUID",
            "LONG128",
            "IPV4",
        ):
            generic_type = GenericDataType.STRING
        elif name_u in ("DATE", "TIMESTAMP"):
            generic_type = GenericDataType.TEMPORAL
        elif "GEOHASH" in name_u and "(" in name_u and ")" in name_u:
            generic_type = GenericDataType.STRING
        return utils.ColumnSpec(
            sqla_type,
            generic_type,
            generic_type == GenericDataType.TEMPORAL,
        )

    @classmethod
    def get_sqla_column_type(
        cls,
        native_type: str | None,
        db_extra: dict[str, Any] | None = None,
        source: utils.ColumnTypeSource = utils.ColumnTypeSource.GET_TABLE,
    ) -> TypeEngine | None:
        """Converts native database type to sqlalchemy column type.
        :param native_type: Native database type
        :param db_extra: The database extra object
        :param source: Type coming from the database table or cursor description
        :return: ColumnSpec object
        """
        resolved = qdbc_types.resolve_type_from_name(native_type)
        return resolved.impl if resolved else None

    @classmethod
    def select_star(  # pylint: disable=too-many-arguments
        cls,
        database: Any,
        table_name: str,
        engine: Engine,
        schema: str | None = None,
        limit: int = 100,
        show_cols: bool = False,
        indent: bool = True,
        latest_partition: bool = True,
        cols: list[dict[str, Any]] | None = None,
    ) -> str:
        """Generate a "SELECT * from table_name" query with appropriate limit.
        :param database: Database instance
        :param table_name: Table name, unquoted
        :param engine: SqlAlchemy Engine instance
        :param schema: Schema, unquoted
        :param limit: limit to impose on query
        :param show_cols: Show columns in query; otherwise use "*"
        :param indent: Add indentation to query
        :param latest_partition: Only query the latest partition
        :param cols: Columns to include in query
        :return: SQL query
        """
        return super().select_star(
            database,
            table_name,
            engine,
            None,
            limit,
            show_cols,
            indent,
            latest_partition,
            cols,
        )

    @classmethod
    def get_allow_cost_estimate(cls, extra: dict[str, Any]) -> bool:
        return False

    @classmethod
    def get_view_names(
        cls,
        database,
        inspector: Inspector,
        schema: str | None,
    ) -> set[str]:
        return set()

    @classmethod
    def get_text_clause(cls, clause: str) -> TextClause:
        """
        SQLAlchemy wrapper to ensure text clauses are escaped properly

        :param clause: string clause with potentially unescaped characters
        :return: text clause with escaped characters
        """
        if cls.allows_escaped_colons:
            clause = clause.replace(":", "\\:")
        return text(remove_public_schema(remove_public_schema(clause)))

    @classmethod
    def execute(  # pylint: disable=unused-argument
        cls,
        cursor: Any,
        query: str,
        **kwargs: Any,
    ) -> None:
        """Execute a SQL query
        :param cursor: Cursor instance
        :param query: Query to execute
        :param kwargs: kwargs to be passed to cursor.execute()
        :return:
        """
        try:
            sql = sql_parse.strip_comments_from_sql(remove_public_schema(query))
            cursor.execute(text(sql))
        except Exception as ex:
            # Log the exception with traceback
            logger.exception("An error occurred: %s", ex)
            raise cls.get_dbapi_mapped_exception(ex) from ex



import sqlalchemy

from .common import quote_identifier

_GEOHASH_BYTE_MAX = 8
_GEOHASH_SHORT_MAX = 16
_GEOHASH_INT_MAX = 32
_GEOHASH_LONG_BITS = 60
_TYPE_CACHE = {
    # key:   '__visit_name__' of the implementor of QDBTypeMixin
    # value: implementor class itself
}


def geohash_type_name(bits):
    if not isinstance(bits, int) or bits < 0 or bits > _GEOHASH_LONG_BITS:
        raise sqlalchemy.exc.ArgumentError(
            f"geohash precision must be int [0, {_GEOHASH_LONG_BITS}]"
        )
    if 0 < bits <= _GEOHASH_BYTE_MAX:
        return "GEOHASH(8b)"
    elif _GEOHASH_BYTE_MAX < bits <= _GEOHASH_SHORT_MAX:
        return "GEOHASH(3c)"
    elif _GEOHASH_SHORT_MAX < bits <= _GEOHASH_INT_MAX:
        return "GEOHASH(6c)"
    return "GEOHASH(12c)"


def geohash_class(bits):
    if not isinstance(bits, int) or bits < 0 or bits > _GEOHASH_LONG_BITS:
        raise sqlalchemy.exc.ArgumentError(
            f"geohash precision must be int [0, {_GEOHASH_LONG_BITS}]"
        )
    if 0 < bits <= _GEOHASH_BYTE_MAX:
        return GeohashByte
    elif _GEOHASH_BYTE_MAX < bits <= _GEOHASH_SHORT_MAX:
        return GeohashShort
    elif _GEOHASH_SHORT_MAX < bits <= _GEOHASH_INT_MAX:
        return GeohashInt
    return GeohashLong


class QDBTypeMixin(sqlalchemy.types.TypeDecorator):
    __visit_name__ = "QDBTypeMixin"
    impl = sqlalchemy.types.String
    cache_ok = True

    @classmethod
    def matches_type_name(cls, type_name):
        return cls if type_name == cls.__visit_name__ else None

    def column_spec(self, column_name):
        return f"{quote_identifier(column_name)} {self.__visit_name__}"

    def compile(self, dialect=None):
        return self.__visit_name__


class Boolean(QDBTypeMixin):
    __visit_name__ = "BOOLEAN"
    impl = sqlalchemy.types.Boolean
    type_code = 1


class Byte(QDBTypeMixin):
    __visit_name__ = "BYTE"
    impl = sqlalchemy.types.Integer
    type_code = 2


class Short(QDBTypeMixin):
    __visit_name__ = "SHORT"
    type_code = 3
    impl = sqlalchemy.types.Integer


class Char(QDBTypeMixin):
    __visit_name__ = "CHAR"
    type_code = 4


class Int(QDBTypeMixin):
    __visit_name__ = "INT"
    type_code = 5
    impl = sqlalchemy.types.Integer


class Long(QDBTypeMixin):
    __visit_name__ = "LONG"
    type_code = 6
    impl = sqlalchemy.types.Integer


class Date(QDBTypeMixin):
    __visit_name__ = "DATE"
    type_code = 7
    impl = sqlalchemy.types.Date


class Timestamp(QDBTypeMixin):
    __visit_name__ = "TIMESTAMP"
    type_code = 8
    impl = sqlalchemy.types.DateTime


class Float(QDBTypeMixin):
    __visit_name__ = "FLOAT"
    type_code = 9
    impl = sqlalchemy.types.Float


class Double(QDBTypeMixin):
    __visit_name__ = "DOUBLE"
    type_code = 10
    impl = sqlalchemy.types.Float


class String(QDBTypeMixin):
    __visit_name__ = "STRING"
    type_code = 11


class Symbol(QDBTypeMixin):
    __visit_name__ = "SYMBOL"
    type_code = 12


class Long256(QDBTypeMixin):
    __visit_name__ = "LONG256"
    type_code = 13


class GeohashByte(QDBTypeMixin):
    __visit_name__ = geohash_type_name(8)
    type_code = 14


class GeohashShort(QDBTypeMixin):
    __visit_name__ = geohash_type_name(16)
    type_code = 15


class GeohashInt(QDBTypeMixin):
    __visit_name__ = geohash_type_name(32)
    type_code = 16


class GeohashLong(QDBTypeMixin):
    __visit_name__ = geohash_type_name(60)
    type_code = 17


class UUID(QDBTypeMixin):
    __visit_name__ = "UUID"
    type_code = 19


class Long128(QDBTypeMixin):
    __visit_name__ = "LONG128"
    type_code = 24


class IPv4(QDBTypeMixin):
    __visit_name__ = "IPV4"
    type_code = 26


QUESTDB_TYPES = [
    Boolean,
    Byte,
    Short,
    Char,
    Int,
    Long,
    Date,
    Timestamp,
    Float,
    Double,
    String,
    Symbol,
    Long256,
    GeohashByte,
    GeohashInt,
    GeohashShort,
    GeohashLong,
    UUID,
    Long128,
    IPv4,
]


def resolve_type_from_name(type_name):
    if not type_name:
        return None
    type_class = _TYPE_CACHE.get(type_name)
    if not type_class:
        for candidate_class in QUESTDB_TYPES:
            type_class = candidate_class.matches_type_name(type_name)
            if type_class:
                _TYPE_CACHE[type_name] = type_class
                break
            elif (
                "GEOHASH" in type_name.upper() and "(" in type_name and ")" in type_name
            ):
                open_p = type_name.index("(")
                close_p = type_name.index(")")
                description = type_name[open_p + 1 : close_p]
                g_size = int(description[:-1])
                if description[-1] in ("C", "c"):
                    g_size *= 5
                type_class = geohash_class(g_size)
                break
    return type_class



import datetime
import json
import os
import time

os.environ.setdefault("SQLALCHEMY_SILENCE_UBER_WARNING", "1")

import questdb_connect as qdbc
from sqlalchemy import Column, MetaData, insert, text
from sqlalchemy.orm import declarative_base

from examples import CONNECTION_ATTRS

Base = declarative_base(metadata=MetaData())
table_name = "all_types"


class MyTable(Base):
    __tablename__ = table_name
    __table_args__ = (
        qdbc.QDBTableEngine(table_name, "col_ts", qdbc.PartitionBy.DAY, is_wal=True),
    )
    col_boolean = Column(qdbc.Boolean)
    col_byte = Column(qdbc.Byte)
    col_short = Column(qdbc.Short)
    col_int = Column(qdbc.Int)
    col_long = Column(qdbc.Long)
    col_float = Column(qdbc.Float)
    col_double = Column(qdbc.Double)
    col_symbol = Column(qdbc.Symbol)
    col_string = Column(qdbc.String)
    col_char = Column(qdbc.Char)
    col_uuid = Column(qdbc.UUID)
    col_date = Column(qdbc.Date)
    col_ts = Column(qdbc.Timestamp, primary_key=True)
    col_geohash = Column(qdbc.GeohashInt)
    col_long256 = Column(qdbc.Long256)


def main():
    # obtain the engine, which we will dispose of at the end in the finally
    engine = qdbc.create_engine(**CONNECTION_ATTRS)
    try:
        # delete any previous existing 'all_types' table
        while True:
            try:
                Base.metadata.drop_all(engine)
                break
            except Exception as see:
                if "Connection refused" in str(see) or (hasattr(see, 'orig') and "Connection refused" in str(see.orig)):
                    print(f"awaiting for QuestDB to start")
                    time.sleep(10)
                else:
                    raise see

        # create the 'all_types' table
        Base.metadata.create_all(engine)

        # connect with QuestDB
        with engine.connect() as conn:
            # insert a fully populated row
            now = datetime.datetime(2023, 4, 22, 18, 10, 10, 765123)
            conn.execute(
                insert(MyTable).values(
                    col_boolean=True,
                    col_byte=8,
                    col_short=12,
                    col_int=13,
                    col_long=14,
                    col_float=15.234,
                    col_double=16.88993244,
                    col_symbol="coconut",
                    col_string="banana",
                    col_char="C",
                    col_uuid="6d5eb038-63d1-4971-8484-30c16e13de5b",
                    col_date=now.date(),
                    col_ts=now,
                    col_geohash="dfvgsj2vptwu",
                    col_long256="0xa3b400fcf6ed707d710d5d4e672305203ed3cc6254d1cefe313e4a465861f42a",
                )
            )
            conn.commit()
            columns = [col.name for col in MyTable.__table__.columns]
            while True:
                rs = conn.execute(text("all_types"))
                if rs.rowcount:
                    print(f"rows: {rs.rowcount}")
                    for row in rs:
                        print(json.dumps(dict(zip(columns, map(str, row))), indent=4))
                    break
    finally:
        if engine:
            engine.dispose()


if __name__ == "__main__":
    main()
